"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import atexit
import os
import sys

import bpy

from PySide2.QtWidgets import QApplication

from .blender_applications import BlenderApplication

# GLOBALS #
TICK = 1.0 / float(os.getenv("BQT_TICK_RATE", "30"))


class QOperator(bpy.types.Operator):
    """
    QOperator is a subclass of the Blender `bpy.types.Operator`
    It instantiates the application if one does not exist already
    """
    bl_idname = "qoperator.global_app"
    bl_label = "Global QApplication"

    def __init__(self):
        super().__init__()
        self._qapp = None


    def execute(self, context) -> set:
        """
        Args:
            context: Blender Context

        Returns:
            set
        """
        self._qapp = instantiate_application()
        return {'PASS_THROUGH'}


# CORE FUNCTIONS #
def instantiate_application() -> BlenderApplication:
    """
    Create an instance of Blender Application

    Returns BlenderApplication: Application Instance

    """
    app = QApplication.instance()
    if not app:
        app = load_os_module()
        bpy.app.timers.register(on_update, persistent=True)

    return app


def load_os_module() -> object:
    """
    Loads the correct OS platform Application Class

    Returns: Instance of BlenderApplication

    """
    operating_system = sys.platform
    if operating_system == 'darwin':
        from .blender_applications.darwin_blender_application import DarwinBlenderApplication
        return DarwinBlenderApplication()
    if operating_system in ['linux', 'linux2']:
        # TODO: LINUX module
        pass
    elif operating_system == 'win32':
        from .blender_applications.win32_blender_application import Win32BlenderApplication
        return Win32BlenderApplication()


def on_update() -> float:
    """
    Checks per Blender timer tick to verify if application should close

    Returns: Tick Rate

    """
    app = QApplication.instance()
    if app.should_close:
        bpy.ops.wm.quit_blender({'window': bpy.context.window_manager.windows[0]}, 'INVOKE_DEFAULT')

    return TICK


@bpy.app.handlers.persistent
def create_global_app(*_args):
    """
    Create global application
    Args:
        *_args:

    Returns:

    """
    if 'startup' in __file__ and not os.getenv('BQT_DISABLE_STARTUP'):
        bpy.ops.qoperator.global_app()


def register():
    """
    Register Blender Operator classes

    Returns: None

    """
    bpy.utils.register_class(QOperator)
    if create_global_app not in bpy.app.handlers.load_post:
        bpy.app.handlers.load_post.append(create_global_app)


def unregister():
    """
    Unregister Blender Operator classes

    Returns: None

    """
    bpy.utils.unregister_class(QOperator)
    if create_global_app in bpy.app.handlers.load_post:
        bpy.app.handlers.load_post.remove(create_global_app)


def on_exit():
    """
    Close BlenderApplication instance on exit

    Returns: None

    """
    app = QApplication.instance()
    if app:
        app.store_window_geometry()
        app.quit()


atexit.register(on_exit)


if __name__ == '__main__':
    try:
        unregister()
    except (ValueError, TypeError) as e:
        print(f"Failed to unregister QOperator: {e}")

    register()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from abc import abstractmethod, abstractstaticmethod, ABCMeta
from pathlib import Path

from PySide2.QtWidgets import QApplication, QWidget
from PySide2.QtGui import QCloseEvent, QIcon, QImage, QPixmap, QWindow
from PySide2.QtCore import QEvent, QObject, QRect, QSettings


class BlenderApplication(QApplication):
    """
    Base Implementation for QT Blender Window Container
    """

    def __init__(self):
        __metaclass__ = ABCMeta
        super().__init__()

        self._stylesheet_filepath = Path(__file__).parent / ".." / "blender_stylesheet.qss"
        self._settings_key_geometry = "Geometry"
        self._settings_key_maximized = "IsMaximized"
        self._settings_key_full_screen = "IsFullScreen"
        self._settings_key_window_group_name = "MainWindow"

        # QApplication
        if self._stylesheet_filepath.exists():
            self.setStyleSheet(self._stylesheet_filepath.read_text())

        QApplication.setWindowIcon(self._get_application_icon())

        # Blender Window
        self._hwnd = self._get_application_hwnd()
        self._blender_window = QWindow.fromWinId(self._hwnd)
        self.blender_widget = QWidget.createWindowContainer(self._blender_window)
        self.blender_widget.setWindowTitle("Blender")

        # Variables
        self.should_close = False

        # Runtime
        self._set_window_geometry()
        self.focusObjectChanged.connect(self._on_focus_object_changed)

    @abstractstaticmethod
    def _get_application_hwnd() -> int:
        """
        This finds the blender application window and collects the
        handler window ID

        Returns int: Handler Window ID
        """

        return -1

    @staticmethod
    def _get_application_icon() -> QIcon:
        """
        This finds the running blender process, extracts the blender icon from the blender.exe file on disk and saves it to the user's temp folder.
        It then creates a QIcon with that data and returns it.

        Returns QIcon: Application Icon
        """

        icon_filepath = Path(__file__).parent / ".." / "blender_icon_16.png"
        icon = QIcon()

        if icon_filepath.exists():
            image = QImage(str(icon_filepath))
            if not image.isNull():
                icon = QIcon(QPixmap().fromImage(image))

        return icon

    @abstractmethod
    def _on_focus_object_changed(self, focus_object: QObject):
        """
        Args:
            focus_object: Object to track focus event
        """

        pass

    def _set_window_geometry(self):
        """
        Loads stored window geometry preferences and applies them to the QWindow.
        .setGeometry() sets the size of the window minus the window frame.
        For this reason it should be set on self.blender_widget.
        """

        settings = QSettings('Tech-Artists.org', 'Blender Qt Wrapper')
        settings.beginGroup(self._settings_key_window_group_name)

        if settings.value(self._settings_key_full_screen, 'false').lower() == 'true':
            self.blender_widget.showFullScreen()
            return

        if settings.value(self._settings_key_maximized, 'false').lower() == 'true':
            self.blender_widget.showMaximized()
            return

        self.blender_widget.setGeometry(settings.value(self._settings_key_geometry, QRect(0, 0, 640, 480)))
        self.blender_widget.show()

        settings.endGroup()
        return

    def notify(self, receiver: QObject, event: QEvent) -> bool:
        """
        Args:
            receiver: Object to recieve event
            event: Event

        Returns: bool
        """

        if isinstance(event, QCloseEvent) and receiver in (self.blender_widget, self._blender_window):
            event.ignore()
            self.store_window_geometry()
            self.should_close = True
            return False

        return super().notify(receiver, event)

    def store_window_geometry(self):
        """
        Stores the current window geometry for the QWindow
        The .geometry() method on QWindow includes the size of the application minus the window frame.
        For that reason the _blender_widget should be used.
        """

        settings = QSettings('Tech-Artists.org', 'Blender Qt Wrapper')
        settings.beginGroup(self._settings_key_window_group_name)
        settings.setValue(self._settings_key_geometry, self.blender_widget.geometry())
        settings.setValue(self._settings_key_maximized, self.blender_widget.isMaximized())
        settings.setValue(self._settings_key_full_screen, self.blender_widget.isFullScreen())
        settings.endGroup()



"""
A demonstration script showing how to parent a QDialog to the Blender main window.
Yes, this could be done with QtWidgets.QMessageBox more easily.
The point is to show how to parent a window to the Blender application.

**Author:**

    Jeff Hanna, jeff.b.hanna@gmail.com, June 1, 2020
"""

from PySide2.QtCore import Qt
from PySide2.QtWidgets import QApplication, QDialog, QHBoxLayout, QLabel


class HelloWorldDialog(QDialog):
    """
    A sample 'Hello World!' QDialog.

    **Arguments:**

        None

    **Keyword Arguments:**

        None

    **Author:**

        Jeff Hanna, jeff.b.hanna@gmail.com, June 1, 2020
    """

    def __init__(self, parent):
        super().__init__(parent, Qt.WindowCloseButtonHint | Qt.WindowSystemMenuHint | Qt.WindowTitleHint)

        self.resize(200, 50)
        self.setWindowTitle('Qt for Python in Blender')

        lbl_hw = QLabel('Hello World!')

        main_layout = QHBoxLayout()
        main_layout.addWidget(lbl_hw)
        self.setLayout(main_layout)


def demo():
    main_window = QApplication.instance().blender_widget
    dlg = HelloWorldDialog(main_window)
    dlg.show()


if __name__ == '__main__':
    demo()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

from contextlib import suppress

import bpy

with suppress(ModuleNotFoundError):
    import win32gui

from PySide2.QtGui import QIcon, QImage, QPixmap
from PySide2.QtCore import QByteArray, QObject

from .blender_application import BlenderApplication


class Win32BlenderApplication(BlenderApplication):
    """
    Windows implementation of BlenderApplication
    """

    def __init__(self):
        super().__init__()


    @staticmethod
    def _get_application_hwnd() -> int:
        """
        This finds the blender application window and collects the
        handler window ID

        Returns int: Handler Window ID
        """

        hwnd = win32gui.FindWindow(None, 'blender')
        return hwnd


    def _on_focus_object_changed(self, focus_object: QObject):
        """
        Args:
            QObject focus_object: Object to track focus change
        """

        if focus_object is self.blender_widget:
            win32gui.SetFocus(self._hwnd)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at https://mozilla.org/MPL/2.0/.

**Generating Distribution Archives:**
https://packaging.python.org/tutorials/packaging-projects/#generating-distribution-archives

** Uploading Distribution Archives:**
https://packaging.python.org/tutorials/packaging-projects/#uploading-the-distribution-archives

**Notes:**
    In order for the Custom Install wrapper to run successfully we have to use the source distribution
    when we are building our archives. This is done by running the command:
        python setup.py sdist

    This creates a *.tar.gz source file that we use to send to PyPi/TestPyPi that will run setup.py
    compared to shipping a wheel, *.whl, which is binary and does not run any post-install logic.

**Contributors:**
    Greg Amato, amatobahn@gmail.com, June 18, 2020.
"""
import os
from pathlib import Path
from setuptools import setup, find_packages
from setuptools.command.install import install
from shutil import copy
from site import getsitepackages
from subprocess import Popen
import sys


class CustomInstall(install):
    """

    """
    def run(self):
        # Run the standard PyPi Copy
        install.run(self)

        # Post install logic:

        # Install pypiwin32 through a separate process since setuptools.setup() can not
        # properly handle the install for an unknown reason compared to a straight pypi install.
        Popen([sys.executable, "-m", "pip", "install", "pypiwin32"])

        # Copy the bqt_startup.py to the blender scripts/startup so that
        # bqt will initialize properly at Blender startup
        startup_file_path = None
        # Find the blender site-packages folder and generate the full path to bqt_startup.py
        for pkg in getsitepackages():
            if 'site-packages' in pkg:
                startup_file_path = Path(pkg) / "bqt" / "dist" / "bqt_startup.py"
                break

        if not os.path.isfile(startup_file_path):
            print("bqt_startup.py was not found. Please manually move bqt_startup.py to scripts/startup")
            return

        # Get the file destination path
        destination_path = Path(sys.executable).parents[2] / "scripts" / "startup"
        # Copy the file to destination path
        copy(startup_file_path, destination_path)

        # Validate that the file was copied properly
        if os.path.isfile(destination_path / "bqt_startup.py"):
            print(f"Successfully copied bqt_startup.py to {destination_path}")
        else:
            print(f"bqt_startup.py was not copied to {destination_path}.")


setup(
    # Metadata
    name="bqt",
    version="0.2.0",
    description="Files to help bootstrap PySide2 with an event loop within Blender.",
    keywords=["Technical", "Art", "TechArt", "TechArtOrg", "Blender", "Qt", "PySide"],
    license="Mozilla Public License 2.0 (MPL 2.0)",
    url="https://github.com/techartorg/bqt",
    author="tech-artists.org",
    # Classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 3.7"
    ],
    # Requirements
    python_requires=">=3.7",
    packages=find_packages(),
    install_requires=["PySide2"],

    # Package Data
    include_package_data=True,
    package_data={
        'bqt': ["*.png", "*.qss", "dist/bqt_startup.py"]
    },

    # Install Wrapper
    cmdclass={'install': CustomInstall}
)
